Imports System.Math
Imports System.Collections.Generic

Module Geometry
    Public g_MinMaxCorners() As Point   ' For debugging.
    Public g_MinMaxBox As Rectangle     ' For debugging.
    Public g_NonCulledPoints() As Point ' For debugging.

    ' Find the points nearest the upper left, upper right,
    ' lower left, and lower right corners.
    Private Sub GetMinMaxCorners(ByVal points As List(Of Point), ByRef ul As Point, ByRef ur As Point, ByRef ll As Point, ByRef lr As Point)
        ' Start with the first point as the solution.
        ul = points(0)
        ur = ul
        ll = ul
        lr = ul

        ' Search the other points.
        For Each pt As Point In points
            If -pt.X - pt.Y > -ul.X - ul.Y Then ul = pt
            If pt.X - pt.Y > ur.X - ur.Y Then ur = pt
            If -pt.X + pt.Y > -ll.X + ll.Y Then ll = pt
            If pt.X + pt.Y > lr.X + lr.Y Then lr = pt
        Next pt

        g_MinMaxCorners = New Point() {ul, ur, lr, ll} ' For debugging.
    End Sub

    ' Find a box that fits inside the MinMax quadrilateral.
    Private Function GetMinMaxBox(ByVal points As List(Of Point)) As Rectangle
        ' Find the MinMax quadrilateral.
        Dim ul As Point
        Dim ur As Point
        Dim ll As Point
        Dim lr As Point
        GetMinMaxCorners(points, ul, ur, ll, lr)

        ' Get the coordinates of a box that lies inside this quadrilateral.
        Dim xmin As Integer
        Dim xmax As Integer
        Dim ymin As Integer
        Dim ymax As Integer
        xmin = ul.X
        ymin = ul.Y

        xmax = ur.X
        If ymin < ur.Y Then ymin = ur.Y

        If xmax > lr.X Then xmax = lr.X
        ymax = lr.Y

        If xmin < ll.X Then xmin = ll.X
        If ymax > ll.Y Then ymax = ll.Y

        Dim result As New Rectangle(xmin, ymin, xmax - xmin, ymax - ymin)
        g_MinMaxBox = result    ' For debugging.
        Return result
    End Function

    ' Cull points out of the convex hull that lie inside the
    ' trapezoid defined by the vertices with smallest and
    ' largest X and Y coordinates.
    ' Return the points that are not culled.
    Private Function HullCull(ByVal points As List(Of Point)) As List(Of Point)
        ' Find a culling box.
        Dim culling_box As Rectangle = GetMinMaxBox(points)

        ' Cull the points.
        Dim results As New List(Of Point)
        For Each pt As Point In points
            ' See if this point lies outside of the culling box.
            If pt.X <= culling_box.Left OrElse _
               pt.X >= culling_box.Right OrElse _
               pt.Y <= culling_box.Top OrElse _
               pt.Y >= culling_box.Bottom _
            Then
                ' This point cannot be culled.
                ' Add it to the results.
                results.Add(pt)
            End If
        Next pt

        ReDim g_NonCulledPoints(0 To results.Count - 1) ' For debugging.
        results.CopyTo(g_NonCulledPoints)               ' For debugging.
        Return results
    End Function

    ' Return the points that make up a polygon's convex hull.
    ' This method leaves the points list unchanged.
    Public Function MakeConvexHull(ByVal points As List(Of Point)) As List(Of Point)
        ' Cull.
        points = HullCull(points)

        ' Find the remaining point with the smallest Y value.
        ' If there's a tie, take the one with the smaller X value.
        Dim best_pt As Point = points(0)
        For Each pt As Point In points
            If (pt.Y < best_pt.Y) OrElse _
                ((pt.Y = best_pt.Y) AndAlso (pt.X < best_pt.X)) _
            Then
                best_pt = pt
            End If
        Next pt

        ' Move this point to the convex hull.
        Dim hull As New List(Of Point)
        hull.Add(best_pt)
        points.Remove(best_pt)

        ' Start wrapping up the other points.
        Dim sweep_angle As Single = 0
        Do
            ' Find the point with smallest AngleValue
            ' from the last point.
            Dim X As Integer = hull(hull.Count - 1).X
            Dim Y As Integer = hull(hull.Count - 1).Y
            best_pt = Nothing
            Dim best_angle As Single = 360

            ' Search the rest of the points.
            For Each pt As Point In points
                Dim test_angle As Single = AngleValue(X, Y, pt.X, pt.Y)
                If (test_angle >= sweep_angle) And _
                   (best_angle > test_angle) _
                Then
                    best_angle = test_angle
                    best_pt = pt
                End If
            Next pt

            ' See if the first point is better.
            ' If so, we are done.
            Dim first_angle As Single = AngleValue(X, Y, hull(0).X, hull(0).Y)
            If (first_angle >= sweep_angle) And _
               (best_angle >= first_angle) _
            Then
                ' The first point is better. We're done.
                Exit Do
            End If

            ' Add the best point to the convex hull.
            hull.Add(best_pt)
            points.Remove(best_pt)

            sweep_angle = best_angle
        Loop

        Return hull
    End Function

    ' Return a number that gives the ordering of angles
    ' WRST horizontal from the point (x1, y1) to (x2, y2).
    ' In other words, AngleValue(x1, y1, x2, y2) is not
    ' the angle, but if:
    '   Angle(x1, y1, x2, y2) > Angle(x1, y1, x2, y2)
    ' then
    '   AngleValue(x1, y1, x2, y2) > AngleValue(x1, y1, x2, y2)
    ' this angle is greater than the angle for another set
    ' of points, then this number for
    '
    ' This function is dy / (dy + dx).
    Private Function AngleValue(ByVal x1 As Integer, ByVal y1 As Integer, ByVal x2 As Integer, ByVal y2 As Integer) As Single
        Dim dx As Single
        Dim dy As Single
        Dim ax As Single
        Dim ay As Single
        Dim t As Single

        dx = x2 - x1
        ax = Abs(dx)
        dy = y2 - y1
        ay = Abs(dy)
        If ax + ay = 0 Then
            ' If the two points are the same, return 360.
            t = 360 / 9
        Else
            t = dy / (ax + ay)
        End If
        If dx < 0 Then
            t = 2 - t
        ElseIf dy < 0 Then
            t = 4 + t
        End If
        AngleValue = t * 90
    End Function
End Module
